////////////////////////////////////////////////////////////////////////////
//
//  DATUM NASTANKA:    12/09/2002
//
//  POSLEDNJA PROMENA: 18/09/2005
//
//  OPIS:              Osnovna forma.
//
////////////////////////////////////////////////////////////////////////////
unit BaseForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  ChildFrm;

type
  TSysButton = (sbNone, sbFictive, sbClose, sbMaximize, sbMinimize);

  TBaseForm = class(TChildForm)
  private
    FActive: Boolean;
    FAuxCaption: WideString;
    FAuxCaptionRect: TRect;
    FAuxCaptionVisible: Boolean;
    FCachePen: TPen;
    FCaptionRect: TRect;
    FNcHitTestState: DWORD;
    FButtonDown: TSysButton;
    FButtonUnderMouse: TSysButton;
    FImage: TBitmap;
    FMoving: Boolean;
    FOldPos: TPoint;
    FOldWindR: TRect;
    FSizing: Boolean;
    FTitleFont: TFont;
    FWindowIsActive: Boolean;
    FWideCaption: WideString;
    function ButtonUnderMouse(const AMousePos: TPoint): TSysButton;
    procedure CalculateCaptionBarRect(var R: TRect; ADoOffset: Boolean = True);
    procedure DrawCaptionBar(DC: HDC);
    procedure DrawSpecialLine(DC: HDC; const ARect: TRect);
    procedure DrawWindowFrame(DC: HDC);
    function GetSysButton(const ASysButton: TSysButton; const AUseRect: TRect): TRect;
    procedure GetCachedObjects;
    procedure SetActive(const Value: Boolean);
    procedure SetAuxCaption(const Value: WideString);
    procedure SetAuxCaptionVisible(const Value: Boolean);
    procedure SetWideCaption(const Value: WideString);
    procedure SizeMovingLoop(const AMousePos: TPoint);
    procedure TitleFontChange(Sender: TObject);
  protected
    procedure CreateWnd; override;
    procedure MyWndProc(var Message: TMessage);
  public
    constructor Create(AOwner: TComponent); override;
    constructor Create(AOwner: TComponent; AParent: TWinControl); override;
    destructor Destroy; override;
  published
    property Active: Boolean read FActive write SetActive;
    property AuxCaption: WideString read FAuxCaption write SetAuxCaption;
    property AuxCaptionVisible: Boolean read FAuxCaptionVisible write SetAuxCaptionVisible default False;
    property TitleFont: TFont read FTitleFont;
    property WideCaption: WideString read FWideCaption write SetWideCaption;
  end;

implementation

{$R *.DFM}

{ TfrmBase }

uses
  Math, JclSysInfo, JwaWinUser, Utils, UtilsEx;

function FillGradient(DC: HDC; ARect: TRect; ColorCount: Integer;
  StartColor, EndColor: TColor): Boolean;
var
  StartRGB: array [0..2] of Byte;
  RGBKoef: array [0..2] of Double;
  Brush: HBRUSH;
  AreaWidth, AreaHeight, i: Integer;
  ColorRect: TRect;
  RectOffset: Double;
begin
  // NOTE:
  // This function is taken from JclGraphics

  Result := False;
  if ColorCount < 1 then
    Exit;
  StartColor := ColorToRGB(StartColor);
  EndColor := ColorToRGB(EndColor);
  StartRGB[0] := GetRValue(StartColor);
  StartRGB[1] := GetGValue(StartColor);
  StartRGB[2] := GetBValue(StartColor);
  RGBKoef[0] := (GetRValue(EndColor) - StartRGB[0]) / ColorCount;
  RGBKoef[1] := (GetGValue(EndColor) - StartRGB[1]) / ColorCount;
  RGBKoef[2] := (GetBValue(EndColor) - StartRGB[2]) / ColorCount;
  AreaWidth := ARect.Right - ARect.Left;
  AreaHeight :=  ARect.Bottom - ARect.Top;
  RectOffset := AreaWidth / ColorCount;

  for i := 0 to ColorCount - 1 do
  begin
    Brush := CreateSolidBrush(RGB(
      StartRGB[0] + Round((i + 1) * RGBKoef[0]),
      StartRGB[1] + Round((i + 1) * RGBKoef[1]),
      StartRGB[2] + Round((i + 1) * RGBKoef[2])));

    SetRect(ColorRect, Round(RectOffset * i), 0, Round(RectOffset * (i + 1)), AreaHeight);

    OffsetRect(ColorRect, ARect.Left, ARect.Top);
    FillRect(DC, ColorRect, Brush);
    DeleteObject(Brush);
  end;
  Result := True;
end;

constructor TBaseForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FAuxCaptionVisible := False;
  FImage := TBitmap.Create;
  Active := False;

  // Setujemo WideCaption na Caption koji je forma imala u vreme kreiranja
  WideCaption := Caption;
end;

constructor TBaseForm.Create(AOwner: TComponent; AParent: TWinControl);
begin
  inherited Create(AOwner, AParent);

  FAuxCaptionVisible := False;
  FImage := TBitmap.Create;
  Active := True;

  // Setujemo WideCaption na Caption koji je forma imala u vreme kreiranja
  WideCaption := Caption;
end;

destructor TBaseForm.Destroy;
begin
  FImage.Free;
  FTitleFont.Free;
  FCachePen.Free;

  inherited Destroy;
end;

function TBaseForm.ButtonUnderMouse(const AMousePos: TPoint): TSysButton;
var
  R: TRect;
begin
  // Testiramo Close button
  Result := sbClose;
  R := GetSysButton(Result, Rect(0, 0, 0, 0));
  if Windows.PtInRect(R, AMousePos) then
    Exit;

  // Maximize button
  Result := sbMaximize;
  R := GetSysButton(Result, Rect(0, 0, 0, 0));
  if Windows.PtInRect(R, AMousePos) then
    Exit;

  // Minimize button
  Result := sbMinimize;
  R := GetSysButton(Result, Rect(0, 0, 0, 0));
  if Windows.PtInRect(R, AMousePos) then
    Exit;

  Result := sbNone;
end;

procedure TBaseForm.CalculateCaptionBarRect(var R: TRect; ADoOffset: Boolean = True);
begin
  // Sracunavamo rect koji zauzima caption bar u koordinatama ekrana
  GetWindowRect(Handle, R);
  if ADoOffset then
    OffsetRect(R, -R.Left, -R.Top);

  case BorderStyle of
    bsToolWindow, bsSingle, bsDialog: InflateRect(R, -GetSystemMetrics(SM_CXFIXEDFRAME), -GetSystemMetrics(SM_CYFIXEDFRAME));
    bsSizeable, bsSizeToolWin       : InflateRect(R, -GetSystemMetrics(SM_CXSIZEFRAME), -GetSystemMetrics(SM_CYSIZEFRAME));
  end;

  if BorderStyle in [bsToolWindow, bsSizeToolWin] then
    R.Bottom := R.Top + GetSystemMetrics(SM_CYSMCAPTION)
  else
    R.Bottom := R.Top + GetSystemMetrics(SM_CYCAPTION);
end;

procedure TBaseForm.DrawCaptionBar(DC: HDC);
var
  R, ImageRect: TRect;
  AuxR, CptR, IconR: TRect;
  ClipRgn: HRGN;

  procedure DrawRoundedEdge(const ACanvas: TCanvas; const R: TRect);
  var
    PenX, PenY, RectWidth, RectHeight, Ellipse: Integer;
  begin
    ACanvas.Brush.Style := bsClear;
    PenX := ACanvas.Pen.Width div 2;
    PenY := PenX;
    RectWidth := R.Right - R.Left - ACanvas.Pen.Width + 1;
    RectHeight := R.Bottom - R.Top - ACanvas.Pen.Width + 1;
    if Canvas.Pen.Width = 0 then
    begin
      Dec(RectWidth);
      Dec(RectHeight);
    end;

    if RectWidth < RectHeight then
      Ellipse := RectWidth
    else
      Ellipse := RectHeight;

    RoundRect(ACanvas.Handle, PenX + R.Left, PenY + R.Top, PenX + R.Left + RectWidth, PenY + R.Top + RectHeight, Ellipse div 4, Ellipse div 4);
    ACanvas.Brush.Style := bsSolid;
  end;

  function DrawSysButtons: TRect;
  var
    BtnRect: TRect;
    Temp: Integer;
  begin
    Result := Rect(0, 0, 0, 0);

    if biSystemMenu in BorderIcons then
    begin
      BtnRect := GetSysButton(sbClose, ImageRect);
      Result := BtnRect;
      if (FButtonUnderMouse = sbClose) and (FButtonDown = FButtonUnderMouse) then
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_SUNKEN, BF_RECT or BF_ADJUST)
      else
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_BUMP, BF_RECT or BF_MONO or BF_ADJUST);

      DrawRoundedEdge(FImage.Canvas, BtnRect);
      FImage.Canvas.Pen.Width := 2;
      FImage.Canvas.Pen.Color := clRed;

      if (FButtonUnderMouse = sbClose) and (FButtonDown = FButtonUnderMouse) then
        OffsetRect(BtnRect, 1, 1);
      Temp := (BtnRect.Right - BtnRect.Left) div 3;

      FImage.Canvas.MoveTo(BtnRect.Left + Temp, BtnRect.Top + Temp);
      FImage.Canvas.LineTo(BtnRect.Right - Temp, BtnRect.Bottom - Temp);

      FImage.Canvas.MoveTo(BtnRect.Left + Temp, BtnRect.Bottom - Temp);
      FImage.Canvas.LineTo(BtnRect.Right - Temp, BtnRect.Top + Temp);
      FImage.Canvas.Pen.Width := 1;
      FImage.Canvas.Pen.Color := clBlack;
    end;

    // Kada je FormBorder style neka od sledecih vrednosti
    // onda ne prikazujemo ostale dugmice
    if BorderStyle in [bsDialog, bsSizeToolWin, bsToolWindow] then
      Exit;

    if biMaximize in BorderIcons then
    begin
      BtnRect := GetSysButton(sbMaximize, ImageRect);
      Result := BtnRect;
      if (FButtonUnderMouse = sbMaximize) and (FButtonDown = FButtonUnderMouse) then
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_SUNKEN, BF_RECT or BF_ADJUST)
      else
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_BUMP, BF_RECT or BF_MONO or BF_ADJUST);

      DrawRoundedEdge(FImage.Canvas, BtnRect);

      FImage.Canvas.Pen.Width := 2;
      FImage.Canvas.Brush.Style := bsClear;

      if (FButtonUnderMouse = sbMaximize) and (FButtonDown = FButtonUnderMouse) then
        OffsetRect(BtnRect, 1, 1);
      {Temp := (BtnRect.Right - BtnRect.Left) div 5;
      Temp := 2;}

      //Rectangle(FImage.Canvas.Handle, BtnRect.Left + Temp, BtnRect.Top + Temp, BtnRect.Right - Temp, BtnRect.Bottom - Temp);
      Rectangle(FImage.Canvas.Handle, BtnRect.Left + 3, BtnRect.Top + 3, BtnRect.Right - 2, BtnRect.Bottom - 2);
      FImage.Canvas.MoveTo(BtnRect.Left + 3, BtnRect.Top + 4);
      FImage.Canvas.LineTo(BtnRect.Right - 3, BtnRect.Top + 4);
      FImage.Canvas.Pen.Width := 1;
    end;

    if biMinimize in BorderIcons then
    begin
      BtnRect := GetSysButton(sbMinimize, ImageRect);
      Result := BtnRect;
      if (FButtonUnderMouse = sbMinimize) and (FButtonDown = FButtonUnderMouse) then
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_SUNKEN, BF_RECT or BF_ADJUST)
      else
        DrawEdge(FImage.Canvas.Handle, BtnRect, EDGE_BUMP, BF_RECT or BF_MONO or BF_ADJUST);

      DrawRoundedEdge(FImage.Canvas, BtnRect);

      FImage.Canvas.Pen.Width := 3;

      if (FButtonUnderMouse = sbMinimize) and (FButtonDown = FButtonUnderMouse) then
        OffsetRect(BtnRect, 1, 1);
      Temp := (BtnRect.Right - BtnRect.Left) div 4;

      FImage.Canvas.MoveTo(BtnRect.Left + Temp, BtnRect.Bottom - Temp);
      FImage.Canvas.LineTo(BtnRect.Right - Temp, BtnRect.Bottom - Temp);
      FImage.Canvas.Pen.Width := 1;
    end;
  end;

  function DrawFormIcon(const ACanvas: TCanvas; const R: TRect): TRect;
  var
    hIco: HICON;
  begin
    Result.Left := R.Left + 2;
    Result.Top := R.Top + (R.Bottom - R.Top - GetSystemMetrics(SM_CYSMICON)) div 2;
    Result.Right := Result.Left + GetSystemMetrics(SM_CXSMICON);
    Result.Bottom := Result.Top + GetSystemMetrics(SM_CYSMICON);

    // Ikonu trazimo sledecim redom: prvo ikonu forme, onda resurs MAINICON
    // i ako ni to ne urodi plodom onda default windows ikona
    if Icon.Handle <> 0 then
      hIco := Icon.Handle
    else begin
      hIco := LoadImage(HInstance, 'MAINICON', IMAGE_ICON, GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), LR_DEFAULTCOLOR or LR_SHARED);
      if hIco = 0 then
        hIco := LoadIcon(0, IDI_APPLICATION);
    end;

    DrawIconEx(ACanvas.Handle, Result.Left, Result.Top, hIco,
      Result.Right - Result.Left, Result.Bottom - Result.Top, 0, 0, DI_NORMAL);
  end;

begin
  if (csDestroying in ComponentState) or (csLoading in ComponentState) or (BorderStyle = bsNone) then
    Exit;

  CalculateCaptionBarRect(R);

  ImageRect := Rect(0, 0, R.Right - R.Left, R.Bottom - R.Top);
  FImage.Width := ImageRect.Right;
  FImage.Height := ImageRect.Bottom;
  // Poslednji red piksela u title baru se crta u boji okvira prozora (clBtnFace najverovatnije)
  // zato sada smanjujemo donju ivicu ImageRecta
  Dec(ImageRect.Bottom);
  DrawSpecialLine(FImage.Canvas.Handle, ImageRect);

  FImage.Canvas.Font.Assign(FTitleFont);

  // Bojimo pozadinu TitleBar-a
  if FWindowIsActive then
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $008F7842, $00D1C482)
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $008F7842, $00F0CAA6) Dugo koriscen kao default
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $008F7842, $08525200)
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $08525200, TColor($80000003))
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, TColor($80000003), $08525200)
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $08525200, TColor($008F7842))
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $008F7842, $08525200)
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $008F7842, $00EFD3C6)
    //FillGradient(FImage.Canvas.Handle, ImageRect, 80, $00C66931, RGB(215, 172, 23))
    FillGradient(FImage.Canvas.Handle, ImageRect, 80, $00C66931, $00F0CAA6)
  else begin
    FImage.Canvas.Brush.Color := TColor($80000003);
    FImage.Canvas.FillRect(ImageRect);
  end;

  // Crtamo ikonu
  if (BorderStyle in [bsSizeable, bsSingle]) or (Icon.Handle <> 0) then
    IconR := DrawFormIcon(FImage.Canvas, ImageRect)
  else
    IconR := Rect(0, 0, 0, 0);

  // Sada krecemo sa crtanjem raznih elemenata caption bara, iscrtavanje vrsimo
  // od desnog do levog dela caption bara

  // CRTAMO SISTEMSKE DUGMICE
  // Povratna vrednost iz DrawSysButtons funkcije je rect poslednjeg
  // iscrtanog dugmeta, AuxCaption crtamo levo od tog recta
  AuxR := DrawSysButtons;
  if IsRectEmpty(AuxR) then
    AuxR := GetSysButton(sbFictive, ImageRect);

  // CRTAMO AUX CAPTION
  if FAuxCaptionVisible then
  begin
    AuxR.Right := AuxR.Left - 2;
    AuxR.Left := AuxR.Left - (AuxR.Left - ImageRect.Left) div 5;
    FAuxCaptionRect := AuxR;

    // FAuxCaptionRsect je u koordinatama CaptionBara, konvertujemo ga u koordinate prozora
    // (dodajemo sirinu i visinu leve i gornje ivice prozora);
    OffsetRect(FAuxCaptionRect, R.Left, R.Top);

    DrawRoundedEdge(FImage.Canvas, AuxR);
    if FAuxCaption <> '' then
    begin
      // Smanjujemo sirinu recta kako bi odvojili tekst od ivice captiona
      InflateRect(AuxR, -3, 0);
      ClipRgn := CreateRectRgnIndirect(AuxR);
      SelectClipRgn(FImage.Canvas.Handle, ClipRgn);
      DeleteObject(ClipRgn);
      FImage.Canvas.Brush.Style := bsClear;
      UtilsEx.DrawTextW(FImage.Canvas.Handle, FAuxCaption, AuxR, DT_SINGLELINE or DT_VCENTER or DT_CENTER or DT_END_ELLIPSIS or DT_NOCLIP);
      SelectClipRgn(FImage.Canvas.Handle, 0);
      Dec(AuxR.Left, 3); // Vracamo smanjenje recta koje smo uradili na pocetku ovog bloka
    end;
  end;

  // CRTAMO CAPTION
  CptR := AuxR;
  CptR.Right := CptR.Left - 2;
  CptR.Left := IconR.Right + 2;

  FCaptionRect := CptR;
  OffsetRect(FCaptionRect, R.Left, R.Top);

  DrawRoundedEdge(FImage.Canvas, CptR);
  if FWideCaption <> '' then
  begin
    // Smanjujemo sirinu recta kako bi odvojili tekst od ivice okvira
    InflateRect(CptR, -R.Left, 0);
    ClipRgn := CreateRectRgnIndirect(CptR);
    SelectClipRgn(FImage.Canvas.Handle, ClipRgn);
    DeleteObject(ClipRgn);
    FImage.Canvas.Brush.Style := bsClear;
    UtilsEx.DrawTextW(FImage.Canvas.Handle, FWideCaption, CptR, DT_SINGLELINE or DT_VCENTER or DT_END_ELLIPSIS or DT_NOCLIP);
    SelectClipRgn(FImage.Canvas.Handle, 0);
  end;

  BitBlt(DC, R.Left, R.Top, FImage.Width, FImage.Height, FImage.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TBaseForm.DrawSpecialLine(DC: HDC; const ARect: TRect);
var
  hOldPen: HGDIOBJ;
begin
  // Crtamo jedan red piksela na dnu recta, ovo je potrebno raditi kod Caption i Menu bara
  hOldPen := SelectObject(DC, FCachePen.Handle);
  MoveToEx(DC, ARect.Left, ARect.Bottom, nil);
  LineTo(DC, ARect.Right, ARect.Bottom);
  SelectObject(DC, hOldPen);
end;

procedure TBaseForm.DrawWindowFrame(DC: HDC);
var
  R: TRect;
  hOldPen, hNewPen: HPEN;
  hOldBrush: HBRUSH;
begin
  if (csDestroying in ComponentState) or (csLoading in ComponentState) or (BorderStyle = bsNone) then
    Exit;
  // Okvir prozora se sastoji iz spoljasnjeg i unutrasnjeg dela, unutrasnji deo je jednobojan
  // i to u boji aktivnog okvira
  GetWindowRect(Handle, R);
  OffsetRect(R, -R.Left, -R.Top);
  DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_ADJUST);

  case BorderStyle of
    bsToolWindow, bsSingle, bsDialog: hNewPen := CreatePen(PS_INSIDEFRAME, GetSystemMetrics(SM_CXFIXEDFRAME) - R.Left, GetSysColor(COLOR_ACTIVEBORDER));
    bsSizeable, bsSizeToolWin       : hNewPen := CreatePen(PS_INSIDEFRAME, GetSystemMetrics(SM_CXSIZEFRAME) - R.Left, GetSysColor(COLOR_ACTIVEBORDER));
    else
      hNewPen := 0;
  end;

  hOldPen := SelectObject(DC, hNewPen);

  hOldBrush := SelectObject(DC, GetStockObject(HOLLOW_BRUSH));
  Rectangle(DC, R.Left, R.Top, R.Right, R.Bottom);
  SelectObject(DC, hOldPen);
  DeleteObject(hNewPen);
  SelectObject(DC, hOldBrush);
end;

function TBaseForm.GetSysButton(const ASysButton: TSysButton; const AUseRect: TRect): TRect;
var
  R: TRect;
  SysBtnWidth, SysBtnHeight: Integer;
begin
  if ASysButton = sbNone then
  begin
    Result := Rect(0, 0, 0, 0);
    Exit;
  end;
  // Prvo pronalazimo Rect za citav CaptionBar
  if not IsRectEmpty(AUseRect) then
    R := AUseRect
  else
    CalculateCaptionBarRect(R, False);

  if BorderStyle in [bsToolWindow, bsSizeToolWin] then
    SysBtnHeight := GetSystemMetrics(SM_CYSMSIZE) - 2
  else
    SysBtnHeight := GetSystemMetrics(SM_CYSIZE) - 2 ;

  SysBtnWidth := SysBtnHeight;

  Result.Top := R.Top + (R.Bottom - R.Top - SysBtnWidth) div 2;
  Result.Bottom := Result.Top + SysBtnHeight;
  Result.Left := R.Right - 1;
  Result.Right := Result.Left + SysBtnWidth;

  if ASysButton = sbFictive then
  begin
    // Ovo je fiktivno dugme koje se koristi kada ni jedno dugme nije vidljivo
    // njegova svrha je da vrati Rect koji cemo koristiti kod crtanja
    // okvira za AuxCaption i Caption
    Result.Left := R.Right;
    Exit;
  end;

  if biSystemMenu in BorderIcons then
  begin
    OffsetRect(Result, -SysBtnWidth - 1, 0);
    if ASysButton = sbClose then
      Exit;
  end;

  // Kada je FormBorder style neka od sledecih vrednosti
  // onda samo prikazujemo Close dugme
  if BorderStyle in [bsDialog, bsSizeToolWin, bsToolWindow] then
  begin
    Result := Rect(0, 0, 0, 0);
    Exit;
  end;

  if biMaximize in BorderIcons then
  begin
    OffsetRect(Result, -SysBtnWidth - 2, 0);
    if ASysButton = sbMaximize then
      Exit;
  end;

  if biMinimize in BorderIcons then
  begin
    OffsetRect(Result, -SysBtnWidth - 2, 0);
    if ASysButton = sbMinimize then
      Exit;
  end;

  Result := Rect(0, 0, 0, 0);
end;

procedure TBaseForm.GetCachedObjects;
var
  NonClientInfo: Windows.TNonClientMetrics;
  FontInfo: Windows.LOGFONT;
begin
  if not Assigned(FTitleFont) then
  begin
    FTitleFont := TFont.Create;
    FCachePen := TPen.Create;
  end;

  // Prvo kreiramo pen objekat koji se koristi za crtanje poslednjeg reda piksela
  // u Caption i Menu baru
  FCachePen.Style := psSolid;
  FCachePen.Width := 1;
  FCachePen.Color := GetSysColor(COLOR_ACTIVEBORDER);

  // Uzimamo da je da je font za caption onaj koji je definisan u sistemu
  NonClientInfo.cbSize := SizeOf(TNonClientMetrics);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(NonClientInfo), @NonClientInfo, 0);

  case BorderStyle of
    bsSizeable, bsSingle, bsDialog: FontInfo := NonClientInfo.lfCaptionFont;
    bsToolWindow, bsSizeToolWin   : FontInfo := NonClientInfo.lfsmCaptionFont;
  end;

  FTitleFont.OnChange := nil;
  FTitleFont.Name := 'Verdana';
  FTitleFont.Height := FontInfo.lfHeight;
  FTitleFont.Charset := FontInfo.lfCharSet;

  if FontInfo.lfWeight <> FW_DONTCARE then
    FTitleFont.Style := FTitleFont.Style + [fsBold];

  if BOOL(FontInfo.lfItalic) then
    FTitleFont.Style := FTitleFont.Style + [fsItalic];

  FTitleFont.OnChange := TitleFontChange;
end;

procedure TBaseForm.MyWndProc(var Message: TMessage);
type
  TGetMenuBarInfo = function (hwnd: HWND; idObject: LongInt; idItem: LongInt; var pmbi: MENUBARINFO): BOOL; stdcall;
var
  MenuRgn: HRGN;
  DC: HDC;
  R: TRect;
  mb: MENUBARINFO;
  _GetMenuBarInfo: TGetMenuBarInfo;
begin
  case Message.Msg of
    WM_NCPAINT, WM_NCACTIVATE, WM_INITMENU:
    begin
      DC := GetWindowDC(Handle);
      try
        case Message.Msg of
          WM_NCPAINT:
          begin
            // Ukoliko prozor ima meni njegovo iscrtavanje prepustamo DefWindowProc s tim sto
            // mu prosledjujemo region koji zauzima sam meni tako postizuci da se
            // ostali non-client elementi prozora ne prepisu.
            _GetMenuBarInfo := GetFunctionAddress(user32, 'GetMenuBarInfo');
            if Assigned(_GetMenuBarInfo) then
            begin
              // GetMenuBarInfo ne postoji na Win95 sistemima stoga koristimo dinamicko linkovanje umesto statickog
              FillChar(mb, SizeOf(MENUBARINFO), 0);
              mb.cbSize := SizeOf(MENUBARINFO);
              _GetMenuBarInfo(Handle, Longint(OBJID_MENU), 0, mb);

              if (mb.hMenu <> 0) and ((Message.wParam = 1) or RectInRegion(HRGN(Message.WParam), mb.rcBar)) then
              begin
                MenuRgn := CreateRectRgnIndirect(mb.rcBar);
                DefWindowProc(Handle, WM_NCPAINT, WPARAM(MenuRgn), Message.LParam);
                DeleteObject(MenuRgn);

                // Iscrtavamo liniju koja razdvaja Meni od Client oblasti prozora
                GetWindowRect(Handle, R);
                OffsetRect(mb.rcBar, -R.Left, -R.Top);
                DrawSpecialLine(DC, mb.rcBar);
              end;
            end;

            // Ako je prosledjen updejt region selektujemo ga, na zalost ne mozemo koristiti
            // standardni WM_NCPAINT mehanizam za prenos HRGN preko WParam jer sistem salje pogresne updejt
            // regione ako se iznad ovog prozora nalazi neki drugi koji se pomera pa onda nastaje
            // zbrlj sa iscrtavanjem. Region je u kordinatama prozora ne kordinatama ekrana
            // SelectClipRgn ocekuje region relativan u odnosu na prozor.
            if Message.LParam <> 1 then
              SelectClipRgn(DC, HRGN(Message.LParam));

            DrawWindowFrame(DC);
            DrawCaptionBar(DC);
          end;

          WM_NCACTIVATE:
          begin
            FButtonDown := sbNone;
            FButtonUnderMouse := sbNone;
            FSizing := False;
            FMoving := False;

            FWindowIsActive := BOOL(Message.wParam);

            // Prozor je aktiviran ili deaktiviran, updejtujmo boju fonta
            if ComponentState * [csDestroying, csLoading] = [] then
            begin
              FTitleFont.OnChange := nil;
              if FWindowIsActive then
                FTitleFont.Color := clCaptionText
              else
                FTitleFont.Color := clInactiveCaptionText;
              FTitleFont.OnChange := TitleFontChange;
            end;

            DrawWindowFrame(DC);
            DrawCaptionBar(DC);
            Message.Result := 1;
          end;

          WM_INITMENU:
          begin
            if GetMenuItemID(GetSystemMenu(Handle, False), 0) = SC_RESTORE then
            begin
              SetMenuDefaultItem(GetSystemMenu(Handle, False), SC_CLOSE, 0);
              EnableMenuItem(GetSystemMenu(Handle, False), SC_MINIMIZE, MF_BYCOMMAND or MF_ENABLED);
              if WindowState <> wsMaximized then
                EnableMenuItem(GetSystemMenu(Handle, False), SC_MAXIMIZE, MF_BYCOMMAND or MF_ENABLED);
            end;

            // Kada se pojavi neki meni onda se iscrtaju default windows sistemski dugmici iscrtaju
            // zato moramo iscrtati nas CaptionBar
            DrawCaptionBar(DC);
          end;
        end;
      finally
        ReleaseDC(Handle, DC);
      end;
    end;

    WM_SETTEXT:
    begin
      // DefWindowProc koja ce na kraju handlovati WM_SETTEXT automatski poziva iscrtavanje
      // captiona prozora i to ne salje WM_NCPAINT vec direktno poziva DrawCaption tako da
      // dolazi do potpuno nepotrebnog i nezeljenog iscrtavanja Captiona i treperanja. Da bi sprecili
      // iscrtavanje Captiona, pre poziva WndProc(Message) (koja na kraju prosledjuje poruku DefWindowProc)
      // saljemo WM_SETREDRAW sa parametrom za onemogucavanje redraw (repaint) citavog prozora,
      // uz napomenu da to radimo samo ako je prozor prikazan. Ako to uvek radimo onda se javlja
      // problem da kada se pozove ShowModal forme i u recimo OnCrete ubaci setovanje captiona,
      // prozor ostane onemogucen za redraw.
      // Kada se zavrsi WndProc onda ponovo pozivamo WM_SETREDRAW ali sada da ukljucimo
      // regularno iscrtavanje prozora.
      if Showing then
        DefWindowProc(Handle, WM_SETREDRAW, 0, 0);

      WndProc(Message);

      if Showing then
        DefWindowProc(Handle, WM_SETREDRAW, 1, 0);
      // Promenio se Caption forme, updejtujemo i WideCaption koji ce okinuti repaint prozora
      WideCaption := Caption;
    end;

    WM_SETTINGCHANGE:
    begin
      WndProc(Message);
      GetCachedObjects;
      Perform(WM_NCPAINT, 1, 0);
    end;

    WM_SETCURSOR:
    begin
      case LoWord(Message.lParam) of
        HTBOTTOM, HTTOP, HTLEFT, HTRIGHT, HTBOTTOMLEFT,
        HTTOPRIGHT, HTBOTTOMRIGHT, HTTOPLEFT:
        begin
          case HiWord(Message.lParam) of
            WM_LBUTTONDOWN:
            begin
              // Zapocinjemo nas size modal loop
              FNcHitTestState := LoWord(Message.LParam);
              FSizing := True;
              Windows.GetCursorPos(FOldPos);
              GetWindowRect(Handle, FOldWindR);
              SetCapture(Handle);
              Message.Result := 1;
            end;

            WM_MOUSEMOVE:
            begin
              case LoWord(Message.lParam) of // Postavljamo kursor za promenu velicine prozora
                HTBOTTOM, HTTOP         : SetCursor(Screen.Cursors[crSizeNS]);
                HTLEFT, HTRIGHT         : SetCursor(Screen.Cursors[crSizeWE]);
                HTBOTTOMLEFT, HTTOPRIGHT: SetCursor(Screen.Cursors[crSizeNESW]);
                HTBOTTOMRIGHT, HTTOPLEFT: SetCursor(Screen.Cursors[crSizeNWSE]);
              end;

              Message.Result := 1;
            end;
          end;
        end;
        else
          WndProc(Message);
      end
    end;

    WM_NCLBUTTONDOWN:
    begin
      case Message.WParam of
        HTCAPTION, HTCLOSE, HTMAXBUTTON, HTMINBUTTON:
        begin
          // Testiramo sta se nalazi ispod misa (sistemsko dugme ili ostatak caption bara)
          FButtonDown := ButtonUnderMouse(SmallPointToPoint(TSmallPoint(Message.LParam)));
          if FButtonDown <> sbNone then
          begin
            FButtonUnderMouse := FButtonDown;

            DC := GetWindowDC(Handle);
            try
              DrawCaptionBar(DC);
            finally
              ReleaseDC(Handle, DC);
            end;

            SetCapture(Handle);
          end
          else
            // Nije pritisnut nijedan od dugmica (a ni ikona, HTSYSMENU parametar),
            // preostalo je pomeranje prozora i aktiviranje prozora
            if not FWindowIsActive then
              BringToFront; // Prozor nije aktivan, aktivirajmo ga

            // Ne mozemo koristiti Windows default moving & sizing modal loop (zezne iscrtavanje)
            // stoga pravimo nas pandan (glavni deo posla se radi u WM_MOUSEMOVE)
            if WindowState <> wsMaximized then
            begin
              FOldPos := SmallPointToPoint(TSmallPoint(Message.LParam));
              SetCapture(Handle);
              FMoving := True;
            end;
        end;

        HTBOTTOM, HTTOP, HTLEFT, HTRIGHT, HTBOTTOMLEFT,
        HTTOPRIGHT,  HTBOTTOMRIGHT, HTTOPLEFT:
        begin
          // Ne dozvoljavamo default Windows ponasanje (iscrtavanje sistemkih dugmica)
        end;
        else
          WndProc(Message);;
      end;
    end;

    WM_SYSCOMMAND:
    begin
      if (Message.wParam and $FFF0 = SC_MINIMIZE) and (fsModal in FormState) then
      begin
        // Ako se zahteva minimizacija bilo koje modalnoj formi, onda minimizujemo
        // citavu aplikaciju (po defaultu Delphi to cini samo za MainForm). SetActiveWindow
        // moramo pozvati jer bez toga se posle minimizovanja non-main forme aplikacija
        // gubi iz Alt + Tab liste
        Application.Minimize;
        SetActiveWindow(Handle);
      end
      else
        WndProc(Message);
    end;

    WM_MOUSEMOVE:
    begin
      // Kako smo capturovali mis onda sve mouse move poruke dolaze ovde
      // WM_NCMOUSEMOVE se ne poziva
      if FButtonDown <> sbNone then
      begin
        FButtonUnderMouse := ButtonUnderMouse(ClientToScreen(SmallPointToPoint(TSmallPoint(Message.LParam))));

        DC := GetWindowDC(Handle);
        try
          DrawCaptionBar(DC);
        finally
          ReleaseDC(Handle, DC);
        end;
      end
      else
        if FSizing or FMoving then
          SizeMovingLoop(ClientToScreen(SmallPointToPoint(TSmallPoint(Message.LParam))))
        else
          WndProc(Message);
    end;

    WM_LBUTTONUP:
    begin
      // Kako smo capturovali mis onda se poruka za otustene taster
      // salje ovde, WM_NCLBUTONUP se ne poziva
      ReleaseCapture;
      // Zavrsavamo nas Size & Moving modal loop
      FSizing := False;
      FMoving := False;
      FNcHitTestState := 0;
      if (FButtonDown = ButtonUnderMouse(ClientToScreen(SmallPointToPoint(TSmallPoint(Message.LParam))))) and (FButtonDown <> sbNone) then
      begin
        // Ako je dugme nad kojim je otpusten mis jednak dugmetu nad kojim je
        // pritisnut mis onda pozivamo odgovarajucu akciju
        case FButtonDown of
          sbClose   : Perform(WM_SYSCOMMAND, SC_CLOSE, Message.LParam);
          sbMaximize:
          begin
            // Ako prozor nije maksimizovan onda ga maksimizujemo
            // Ako jeste, onda ga Restauriramo
            if WindowState = wsMaximized then
              Perform(WM_SYSCOMMAND, SC_RESTORE, Message.LParam)
            else
              Perform(WM_SYSCOMMAND, SC_MAXIMIZE, Message.LParam);
            end;

          sbMinimize:
          begin
            if WindowState = wsMinimized then
              Perform(WM_SYSCOMMAND, SC_RESTORE, Message.lParam)
            else
              Perform(WM_SYSCOMMAND, SC_MINIMIZE, Message.LParam);
          end;
        end;

        FButtonDown := sbNone;
        FButtonUnderMouse := sbNone;
        DC := GetWindowDC(Handle);
        try
          DrawCaptionBar(DC);
        finally
          ReleaseDC(Handle, DC);
        end;
      end
      else begin
        WndProc(Message);
        FButtonDown := sbNone;
        FButtonUnderMouse := sbNone;
      end;
    end;
    else
      WndProc(Message);
  end;
end;

procedure TBaseForm.SetActive(const Value: Boolean);
begin
  if FActive <> Value then
  begin
    FActive := Value;
    // Win95 i WinNT nisu podrzani. Izuzetak je WinNT4 SP6.
    if FActive and not (IsWin95 or (IsWinNT4 and (GetWindowsServicePackVersion < 6))) then
      WindowProc := MyWndProc
    else
      WindowProc := WndProc;

    Perform(WM_NCPAINT, 1, 0);
  end;
end;

procedure TBaseForm.SetAuxCaption(const Value: WideString);
var
  AuxRgn: HRGN;
begin
  if FAuxCaption <> Value then
  begin
    FAuxCaption := Value;

    if FActive and FAuxCaptionVisible then
    begin
      // Kreiramo clipping region tako da se promeni samo oblast AuxCaptiona
      AuxRgn := CreateRectRgnIndirect(FAuxCaptionRect);
      Perform(WM_NCPAINT, 1, LParam(AuxRgn)); // Za objasnjenje zasto koristimo LParam a ne WParam
      DeleteObject(AuxRgn);                   // videti MyWndProc i WM_NCPAINT
    end;
  end;
end;

procedure TBaseForm.SetAuxCaptionVisible(const Value: Boolean);
begin
  if FAuxCaptionVisible <> Value then
  begin
    FAuxCaptionVisible := Value;
    Perform(WM_NCPAINT, 1, 0);
  end;
end;

procedure TBaseForm.SetWideCaption(const Value: WideString);
var
  CaptionRgn: HRGN;
begin
  if FWideCaption <> Value then
  begin
    FWideCaption := Value;

    if FActive then
    begin
      CaptionRgn := CreateRectRgnIndirect(FCaptionRect);
      Perform(WM_NCPAINT, 1, LParam(CaptionRgn));
      DeleteObject(CaptionRgn);
    end
    else begin
      // Ako nije ukljuceno nase iscrtavanje prozora onda se mora koristiti standardni Caption
      // za prikazivanje naslova forme, pa stoga dodeljujemo vrednost Captionu ako to vec nije
      if Caption <> FWideCaption then
        Caption := FWideCaption;
    end;
  end;
end;

procedure TBaseForm.SizeMovingLoop(const AMousePos: TPoint);
var
  L, T, W, H, Temp: Integer;
begin
  L := 0;
  T := 0;
  W := 0;
  H := 0;

  if FMoving then
  begin
    L := FOldPos.x - AMousePos.x;
    T := FOldPos.y - AMousePos.y;
  end
  else begin
    if FNcHitTestState in [HTBOTTOM, HTBOTTOMLEFT, HTBOTTOMRIGHT] then
    begin
      if (Height > GetSystemMetrics(SM_CYMIN)) or (AMousePos.y > Top + Height) then
      begin
        H := FOldPos.y - AMousePos.y;

        // Osiguravamo da donja ivica bude u visini misa
        if Top + Height - H > AMousePos.Y then
          H := Top + Height - AMousePos.y;
      end;
    end;

    if FNcHitTestState in [HTRIGHT, HTBOTTOMRIGHT, HTTOPRIGHT] then
    begin
      if (Width > GetSystemMetrics(SM_CXMIN)) or (AMousePos.x > Left + Width) then
      begin
        W := FOldPos.x - AMousePos.x;

        // Osiguravamo da desna ivica prozora ne dodje iza misa
       if Left + Width - W > AMousePos.x then
          W := Left + Width - AMousePos.x;
      end;
    end;

    if FNcHitTestState in [HTLEFT, HTBOTTOMLEFT, HTTOPLEFT] then
    begin
      if (Width > GetSystemMetrics(SM_CXMIN)) or (AMousePos.x < Left) then
      begin
        L := FOldPos.x - AMousePos.x;
        W := -L;

        if Left < AMousePos.x then
        begin
          // Obezbedjujemo da leva ivica prozora bude poravnata sa misem
          L := Left - AMousePos.x;
          W := -L;
        end;

        Temp := Max(GetSystemMetrics(SM_CXMIN), Constraints.MinWidth);

        if Width - W < Temp then
        begin
          // Ako bi nova sirina trebala da bude manja od minimalne sirine prozora
          // onda je podesavamo tako da bude jednaka minimalnoj sirini
          W := Width - Temp;
          L := -W;
        end;
      end;
    end;

    if FNcHitTestState in [HTTOP, HTTOPLEFT, HTTOPRIGHT] then
    begin
      if (Height > GetSystemMetrics(SM_CYMIN)) or (AMousePos.y < Top) then
      begin
        T := FOldPos.y - AMousePos.y;
        H := -T;

        if Top - T < AMousePos.y then
        begin
          // Obezbedjujemo da gornja ivica prozora bude u visini misa
          T := Top - AMousePos.y;
          H := -T;
        end;

        Temp := Max(GetSystemMetrics(SM_CYMIN), Constraints.MinHeight);

        if Height - H < Temp then
        begin
          // Ako bi nova visina trebala da bude manja od minimalne visine
          // onda je podesavamo tako da bude jednaka minimalnoj visini
          H := Height - Temp;
          T := -H;
        end;
      end;
    end;
  end;

  MoveWindow(Handle, Left - L, Top - T, Width - W, Height - H, True);
  FOldPos := AMousePos;
end;

procedure TBaseForm.TitleFontChange(Sender: TObject);
begin
  Perform(WM_NCPAINT, 1, 0); // Promenjen je font, repaintujemo prozor
end;

procedure TBaseForm.CreateWnd;
begin
  inherited CreateWnd;

  // Ova metoda se poziva kada se kreira prozor bilo po prvi put bilo posle promene nekog
  // parametra (npr. BorderStyle), saznajemo informacije koje su nam potrebne za dalji rad
  GetCachedObjects;
end;

end.
